from os.path import relpath, splitext, basename, join, dirname, exists
from urllib.parse import urljoin
from collections import namedtuple
from .settings import Settings


class Bundle:
    def __init__(self, path: str, root: str, settings: Settings) -> None:
        self.relpath = relpath(path, root)

        root_relpath = splitext(self.relpath)[0]
        socket_relpath = root_relpath + '-bundler.sock'

        out_relpath = root_relpath + '.js'
        out_file = basename(out_relpath)
        if settings.env['NODE_ENV'] == 'production':
            hashed_out_relpath = root_relpath
            hashed_out_relpath += '-' + settings.build_hash + '.js'
        else:
            hashed_out_relpath = out_relpath
        hashed_out_file = basename(hashed_out_relpath)

        out_dir = dirname(self.relpath)
        self.url = urljoin(settings.static_url, out_dir)
        self.component_relpath = relpath(path, settings.entries_dir)

        self.server = Env(
            entry=join(settings.entries_dir, 'server.js'),
            out_file=out_file,
            out_dir=join(settings.bundles_dir, out_dir),
            script_relpath=relpath(
                settings.server_script, settings.entries_dir),
            socket=join(settings.sockets_dir, 'client', socket_relpath),
            cache_dir=join(settings.cache_dir, 'server')
        )
        self.client = Env(
            entry=join(settings.entries_dir, 'client.js'),
            out_file=hashed_out_file,
            out_dir=join(settings.dist_dir, out_dir),
            script_relpath=relpath(
                settings.client_script, settings.entries_dir),
            socket=join(settings.sockets_dir, 'server', socket_relpath),
            cache_dir=join(settings.cache_dir, 'client')
        )

        self.script = urljoin(settings.static_url, hashed_out_relpath)
        stylesheet_relpath = splitext(hashed_out_relpath)[0] + '.css'
        if exists(join(settings.bundles_dir, stylesheet_relpath)):
            self.stylesheet = urljoin(settings.static_url, stylesheet_relpath)
        else:
            self.stylesheet = ''


Env = namedtuple('Env', [
    'entry',
    'out_file',
    'out_dir',
    'script_relpath',
    'socket',
    'cache_dir'
])



from typing import List
from threading import Thread


class Worker():
    _state = {}  # type: dict

    def __init__(self, threads: List[Thread] = None) -> None:
        self.__dict__ = self._state
        if threads is not None:
            self._threads = threads

    @property
    def threads(self) -> List[Thread]:
        if hasattr(self, '_threads'):
            return self._threads
        else:
            raise EnvironmentError(
                'Server side rendering improperly configured. Did you forget '
                'to include the template engine in your Django settings?'
            )

    def run(self) -> None:
        for thread in self.threads:
            thread.start()
        for thread in self.threads:
            thread.join()



from django.template import TemplateSyntaxError
from django.http import HttpRequest
from .bundle import Bundle
from .server import Server


class Template:
    def __init__(self, bundle: Bundle, server: Server) -> None:
        self.bundle = bundle
        self.server = server

    def render(self, context: dict = None, request: HttpRequest = None) -> str:
        try:
            return self.server.render(self.bundle, context)
        except Exception as exception:
            raise TemplateSyntaxError(exception)



import json
from os import makedirs, remove, setsid, environ
from os.path import splitext, exists, dirname
from urllib.parse import quote_plus, urlencode
from subprocess import Popen, PIPE
from threading import Thread
from time import sleep
from typing import Callable, Dict
from requests import codes
from requests_unixsocket import Session
from .utils import wait_for_signal, read_output
from .bundle import Bundle
from .settings import Settings


class Bundler:
    session = Session()

    def __init__(self, bundle: Bundle, settings: Settings) -> None:
        self.bundle = bundle
        self.settings = settings

    def bundle_all(self) -> None:
        server_thread = Thread(target=self.bundle_server)
        client_thread = Thread(target=self.bundle_client)
        server_thread.start()
        client_thread.start()
        server_thread.join()
        client_thread.join()

    def bundle_server(self) -> None:
        if self.settings.env['NODE_ENV'] == 'production':
            self._bundle_server()
        else:
            self._watch(self.bundle.server.socket, self._bundle_server)

    def bundle_client(self) -> None:
        if self.settings.env['NODE_ENV'] == 'production':
            self._bundle_client()
        else:
            self._watch(self.bundle.client.socket, self._bundle_client)

    def _bundle_server(self) -> None:
        self._bundle({
            'SOCKET': self.bundle.server.socket,
            'COMPONENT': self.bundle.component_relpath,
            'SCRIPT': self.bundle.server.script_relpath,
            'PARCEL_OPTIONS': json.dumps({
                'entry': self.bundle.server.entry,
                'config': {
                    'outDir': self.bundle.server.out_dir,
                    'outFile': self.bundle.server.out_file,
                    'cache': self.settings.cache,
                    'cacheDir': self.bundle.server.cache_dir,
                    'sourceMaps': False,
                }
            })
        })

    def _bundle_client(self) -> None:
        self._bundle({
            'SOCKET': self.bundle.client.socket,
            'COMPONENT': self.bundle.component_relpath,
            'SCRIPT': self.bundle.client.script_relpath,
            'PARCEL_OPTIONS': json.dumps({
                'entry': self.bundle.client.entry,
                'config': {
                    'outDir': self.bundle.client.out_dir,
                    'outFile': self.bundle.client.out_file,
                    'cache': self.settings.cache,
                    'cacheDir': self.bundle.client.cache_dir,
                    'publicUrl': self.bundle.url
                }
            })
        })

    def _watch(self, socket: str, watcher: Callable) -> None:
        makedirs(dirname(socket), exist_ok=True)
        if exists(socket):
            if self._reconnect(socket):
                return
            remove(socket)
        watcher()
        self._connect(socket)

    def _poll(self, socket: str) -> None:
        url = self._get_url(socket)
        while True:
            response = self.session.get(url)
            message = response.content.strip(b' ')
            if message:
                print(message.decode('utf-8')[:-1])
            else:
                sleep(0.1)

    def _get_url(self, socket: str) -> str:
        return 'http+unix://' + quote_plus(socket) + '?' + urlencode({
            'pid': self.settings.env['DJANGO_PID']
        })

    def _connect(self, socket: str) -> None:
        Thread(target=self._poll, daemon=True, args=[socket]).start()

    def _reconnect(self, socket: str) -> bool:
        try:
            url = self._get_url(socket)
            response = self.session.get(url)
            connection_exists = response.status_code == codes.ok
            if connection_exists:
                self._connect(socket)
            return connection_exists
        except:
            return False

    def _bundle(self, env: Dict[str, str] = {}) -> None:
        process = Popen([
            'node', self.settings.bundler
        ], stdout=PIPE, stderr=PIPE, preexec_fn=setsid, env={
            **environ,
            **self.settings.env,
            **env
        })
        if self.settings.env['NODE_ENV'] == 'production':
            read_output(process)
        else:
            stdout_thread = Thread(target=wait_for_signal, daemon=True, args=[
                process.stdout, self.settings.env['SIGNAL']
            ])
            stderr_thread = Thread(target=wait_for_signal, daemon=True, args=[
                process.stderr, self.settings.env['SIGNAL']
            ])
            stdout_thread.start()
            stderr_thread.start()
            stdout_thread.join()
            stderr_thread.join()



from os import getpid
from os.path import join
from urllib.parse import urljoin
from platform import system
from tempfile import gettempdir
from uuid import uuid1
from django.conf import settings
from django.utils.module_loading import import_string
from django.core.serializers.json import DjangoJSONEncoder


class Settings:
    build_hash = str(uuid1())
    base_dir = join(settings.BASE_DIR, '.ssr')
    temp_dir = '/tmp' if system() == 'Darwin' else gettempdir()
    scripts_dir = join(base_dir, 'scripts')
    entries_dir = join(scripts_dir, 'entries')
    cache_dir = join(base_dir, 'cache')
    bundles_dir = join(base_dir, 'bundles')
    sockets_dir = join(temp_dir, 'ssr')
    socket = join(sockets_dir, 'renderer.sock')
    server = join(scripts_dir, 'server.mjs')
    bundler = join(scripts_dir, 'bundler.mjs')

    def __init__(self, options: dict) -> None:
        env = {
            'NODE_ENV': 'development' if settings.DEBUG else 'production',
            'NODE_OPTIONS': '--experimental-modules --no-warnings',
            'WORKER_TTL': 1000
        }
        if 'env' in options:
            env = {
                **env,
                **options.pop('env')
            }
        self.env = {
            **env,  # type: ignore
            'WORKER_TTL': str(env['WORKER_TTL']),
            'SIGNAL': self.build_hash,
            'DJANGO_PID': str(getpid())
        }

        if 'json_encoder' in options:
            self.json_encoder = import_string(options.pop('json_encoder'))
        else:
            self.json_encoder = DjangoJSONEncoder

        if 'cache' in options:
            self.cache = options.pop('cache')
        else:
            self.cache = True

        if 'output_dirname' in options:
            output_dirname = options.pop('output_dirname')
        else:
            output_dirname = 'dist/'

        self.static_url = urljoin(settings.STATIC_URL, output_dirname)
        self.dist_dir = join(self.base_dir, 'static', output_dirname)

        scripts = {
            'server': join(self.scripts_dir, 'react', 'server.js'),
            'client': join(self.scripts_dir, 'react', 'client.js'),
        }
        if 'scripts' in options:
            scripts = {
                **scripts,
                **options.pop('scripts')
            }
        self.server_script = scripts['server']
        self.client_script = scripts['client']



from os.path import exists, join, dirname, abspath
from shutil import rmtree, copytree
from glob import glob
from threading import Thread
from django.template.backends.base import BaseEngine
from django.template import TemplateDoesNotExist, TemplateSyntaxError
from typing import Dict
from .settings import Settings
from .server import Server
from .bundle import Bundle
from .bundler import Bundler
from .worker import Worker
from .template import Template


class Components(BaseEngine):
    app_dirname = 'bundles'
    templates = {}  # type: Dict[str, Template]

    def __init__(self, params: dict) -> None:
        params = params.copy()
        options = params.pop('OPTIONS').copy()
        super().__init__(params)

        if 'extensions' in options:
            extensions = options.pop('extensions')
        else:
            extensions = ['js', 'jsx', 'ts', 'tsx']

        settings = Settings(options)
        server = Server(settings)
        server_thread = Thread(target=server.run)

        if exists(settings.scripts_dir):
            rmtree(settings.scripts_dir, ignore_errors=True)
        scripts_dir = join(dirname(abspath(__file__)), 'scripts')
        copytree(scripts_dir, settings.scripts_dir)

        threads = [server_thread]
        for template_dir in self.template_dirs:
            for extension in extensions:
                pattern = join(template_dir, '**', '*.' + extension)
                templates = glob(pattern, recursive=True)
                for template in templates:
                    bundle = Bundle(template, template_dir, settings)
                    bundler = Bundler(bundle, settings)

                    template = Template(bundle, server)
                    self.templates[bundle.relpath] = template

                    thread = Thread(target=bundler.bundle_all)
                    threads.append(thread)
        Worker(threads)

    def get_template(self, template_name: str) -> Template:
        if template_name in self.templates:
            return self.templates[template_name]
        else:
            raise TemplateDoesNotExist('Component not found', backend=self)



import json
from os import makedirs, remove, setsid, environ
from os.path import join, dirname, exists
from urllib.parse import quote_plus, urlencode
from threading import Thread
from subprocess import Popen, PIPE
from requests import codes
from requests_unixsocket import Session
from .settings import Settings
from .utils import wait_for_signal
from .bundle import Bundle


class Server:
    socket_name = 'renderer.sock'
    session = Session()

    def __init__(self, settings: Settings) -> None:
        self.settings = settings
        self.socket = join(self.settings.sockets_dir, self.socket_name)

    def get_url(self, path: str = '', params: dict = {}) -> str:
        host = quote_plus(self.socket) + path
        querystring = urlencode(params)
        return 'http+unix://' + host + '?' + querystring

    @property
    def exists(self) -> bool:
        url = self.get_url('', {
            'pid': self.settings.env['DJANGO_PID']
        })
        try:
            response = self.session.get(url)
            if response.status_code == codes.ok:
                return True
        except:
            pass
        return False

    def run(self) -> None:
        if self.exists:
            return

        makedirs(dirname(self.socket), exist_ok=True)
        if exists(self.socket):
            remove(self.socket)

        process = Popen([
            'node', self.settings.server
        ], stdout=PIPE, stderr=PIPE, preexec_fn=setsid, env={
            **environ,
            **self.settings.env,
            'SOCKET': self.socket,
        })
        stdout_thread = Thread(target=wait_for_signal, args=[
            process.stdout, self.settings.env['SIGNAL']
        ])
        stderr_thread = Thread(target=wait_for_signal, args=[
            process.stderr, self.settings.env['SIGNAL']
        ])
        stdout_thread.start()
        stderr_thread.start()
        stdout_thread.join()
        stderr_thread.join()

    def render(self, bundle: Bundle, props: dict = None) -> str:
        url = self.get_url('/render', {
            'bundle': join(bundle.server.out_dir, bundle.server.out_file),
            'props': json.dumps(props, cls=self.settings.json_encoder),
            'script': bundle.script,
            'stylesheet': bundle.stylesheet
        })
        response = self.session.get(url)
        if response.status_code == codes.ok:
            return response.text
        else:
            raise EnvironmentError(response.text)
